#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open("README.md") as readme_file:
    readme = readme_file.read()

with open("CHANGELOG.md") as history_file:
    history = history_file.read()

requirements = [
    "kubernetes>=9.0.0",
    "dill>=0.2.8.2",
    "deprecation>=2.0.6",
    "Click>=6.0",
    "tabulate>=0.8.5",
    "dill==0.2.8.2",
    "grpcio==1.26.0",
    "pyhelm==2.14.5",
    "appdirs==1.4.3",
    "google-api-python-client==1.7.11",
    "google-auth==1.14.0",
    "google-cloud==0.34.0",
    "google-cloud-container==0.3.0",
    "oauth2client==4.1.2",
    "torchtext==0.5.0",
    "spacy==2.2.3",
    "sklearn==0.0",
]

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest>=3",
    "pytest-mock",
    "deprecation==2.0.6",
    "freezegun==0.3.12",
    "black==19.10b0",
    "pytest-black==0.3.8",
    "pre-commit",
]

setup(
    author="Ralf Grubenmann",
    author_email="ralf.grubenmann@epfl.ch",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    description="A public and reproducible collection of reference implementations and benchmark suite for distributed machine learning systems.",
    entry_points={"console_scripts": ["mlbench=mlbench_core.cli:cli",],},
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="mlbench",
    name="mlbench_core",
    packages=find_packages(),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/mlbench/mlbench_core",
    version="2.4.0-dev66",
    zip_safe=False,
)



"""Contains definitions for Residual Networks.

Residual networks were originally proposed in :cite:`he2016deep` . Then they improve the :cite:`he2016identity`
Here we refer to the settings in :cite:`he2016deep` as `v1` and :cite:`he2016identity`  as `v2`.

Since `torchvision resnet <https://github.com/pytorch/vision/blob/master/torchvision/models/resnet.py>`_
has already implemented.

* ResNet-18
* ResNet-34
* ResNet-50
* ResNet-101
* ResNet-152

for image net. Here we only implemented the remaining models

* ResNet-20
* ResNet-32
* ResNet-44
* ResNet-56

for CIFAR-10 dataset. Besides, their implementation uses projection shortcut by default.

"""
from mlbench_core.utils.pytorch.helpers import convert_dtype

import torch
import torch.nn as nn
from torch.nn import functional as F


_DEFAULT_RESNETCIFAR_VERSION = 1


def batch_norm(num_features):
    """Create a batch normalization layer.

    See the Disclaimers in Kaiming's
    `repository <https://github.com/KaimingHe/deep-residual-networks/tree/a7026cb6d478e131b765b898c312e25f9f6dc031>`_.

    * compute the mean and variance on a sufficiently large traing batch instead of moving average;
    * learn gamma and beta in affine function.

    :param num_features: number of features passed to batch normalization
    :type num_features: int
    """
    return nn.BatchNorm2d(
        num_features=num_features,
        eps=1e-05,
        momentum=0.1,
        affine=True,
        track_running_stats=True,
    )


def conv3x3(in_channels, out_channels, stride=1):
    """3x3 convolution with padding."""
    return nn.Conv2d(
        in_channels, out_channels, kernel_size=3, stride=stride, padding=1, bias=False
    )


class BasicBlockV1(nn.Module):
    """The basic block in :cite:`he2016deep` is used for shallower ResNets.

    The activation functions (ReLU and BN) are viewed as post-activation of the weight layer.

    .. note::
        This class is similar to `BasicBlock` in
        `resnet <https://github.com/pytorch/vision/blob/master/torchvision/models/resnet.py>`_.
        but with different nn.BatchNorm2d configuration.
    """

    def __init__(self, in_channels, out_channels, stride=1, downsample=None):
        """
        :param in_channels: input channels
        :type in_channels: int
        :param out_channels: output channels
        :type out_channels: int
        :param stride: stride of the first layer, defaults to 1
        :type stride: int, optional
        :param downsample: projection identity map or no downsample, defaults to None
        :type downsample: nn.module or None, optional
        """
        super(BasicBlockV1, self).__init__()

        self.conv1 = conv3x3(in_channels, out_channels, stride)
        self.bn1 = batch_norm(out_channels)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(out_channels, out_channels)
        self.bn2 = batch_norm(out_channels)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = self.downsample(x) if self.downsample is not None else x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)

        # Shortcut connection.
        out += residual
        out = self.relu(out)
        return out


class BasicBlockV2(nn.Module):
    """The basic block in :cite:`he2016identity` is used for shallower ResNets.

    The activation functions (ReLU and BN) are viewed as pre-activation of the weight layer.
    """

    def __init__(self, in_channels, out_channels, stride=1, downsample=None):
        """
        :param in_channels: input channels
        :type in_channels: int
        :param out_channels: output channels
        :type out_channels: int
        :param stride: stride of the first layer, defaults to 1
        :type stride: int, optional
        :param downsample: projection identity map or no downsample, defaults to None
        :type downsample: nn.module or None, optional
        """
        super(BasicBlockV2, self).__init__()

        self.bn1 = batch_norm(in_channels)
        self.relu = nn.ReLU(inplace=True)
        self.conv1 = conv3x3(in_channels, out_channels, stride)
        self.bn2 = batch_norm(out_channels)
        self.conv2 = conv3x3(out_channels, out_channels)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = self.downsample(x) if self.downsample is not None else x

        out = self.bn1(x)
        out = self.relu(out)
        out = self.conv1(out)
        out = self.bn2(out)
        out = self.relu(out)
        out = self.conv2(out)

        # Shortcut connection.
        out += residual
        return out


# class BottleneckBlockV1(nn.Module):
#     """Bottleneck building block proposed in :cite:`he2016deep` (post-activation)."""
#     pass


# class BottleneckBlockV2(nn.Module):
#     """Bottleneck building block proposed in :cite:`he2016identity` (post-activation)."""
#     pass


class ResNetCIFAR(nn.Module):
    """Basic ResNet implementation.

    Args:
        resnet_size (int): Number of layers
        bottleneck (bool): Whether to use a bottleneck layer (``Not Implemented``)
        num_classes (int): Number of output classes
        version (int): Resnet version (1 or 2). Default: ``1``
    """

    def __init__(
        self, resnet_size, bottleneck, num_classes, version=_DEFAULT_RESNETCIFAR_VERSION
    ):
        super(ResNetCIFAR, self).__init__()

        if resnet_size % 6 != 2:
            raise ValueError(
                "The resnet_size should be (6 * num_blocks + 2). Got {}.".format(
                    resnet_size
                )
            )

        num_blocks = (resnet_size - 2) // 6

        if version not in (1, 2):
            raise ValueError("Resnet version should be 1 or 2, got {}.".format(version))

        if bottleneck:
            raise NotImplementedError
        else:
            if version == 1:
                block = BasicBlockV1
            elif version == 2:
                block = BasicBlockV2
            else:
                raise NotImplementedError

        # The first layer
        if version == 1 or version == 2:
            self.prep = nn.Sequential(
                conv3x3(in_channels=3, out_channels=16, stride=1),
                batch_norm(num_features=16),
                nn.ReLU(),
            )
        else:
            raise NotImplementedError

        # 6n layers
        self.conv_1 = self._make_layer(
            block, in_channels=16, out_channels=16, num_blocks=num_blocks, init_stride=1
        )
        self.conv_2 = self._make_layer(
            block, in_channels=16, out_channels=32, num_blocks=num_blocks, init_stride=2
        )
        self.conv_3 = self._make_layer(
            block, in_channels=32, out_channels=64, num_blocks=num_blocks, init_stride=2
        )

        # Add an average pooling layer:
        # the output of conv_3 has shape H=W=8
        # the output average pooling will be (batch_size, channels, 1, 1)
        self.avgpool = nn.AvgPool2d(8, stride=8)

        self.classifier = nn.Linear(in_features=64, out_features=num_classes, bias=True)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def _make_layer(self, block, in_channels, out_channels, num_blocks, init_stride=1):
        """Create a block of 2*n depth.

        .. note::
            In :cite:`he2016deep` there are two types of shortcuts: identity and projection. Here we use the following:
            * identity shortcut for same number of channels
            * projection shortcut for increasing number of channels

        """
        # by the design of ResNet, if the init_stride > 1, then out_channels > in_channels.
        # For project shortcut, the extra channels are created using a 1*1 convolution.
        # For identity shortcut, zeros are padded.
        if init_stride == 1:
            downsample = None
        else:
            # Use projection when the dimension of channel increases.
            downsample = nn.Sequential(
                nn.Conv2d(
                    in_channels,
                    out_channels,
                    kernel_size=1,
                    stride=init_stride,
                    bias=False,
                ),
                batch_norm(num_features=out_channels),
            )

        # Maybe use downsample in the first block.
        layers = [
            block(in_channels, out_channels, stride=init_stride, downsample=downsample)
        ]
        for _ in range(1, num_blocks):
            layers.append(block(out_channels, out_channels))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.prep(x)
        x = self.conv_1(x)
        x = self.conv_2(x)
        x = self.conv_3(x)
        x = self.avgpool(x)

        # The plane has shape (1, 1)
        x = x.view(x.size(0), -1)
        x = self.classifier(x)
        return x


""" Version 2 of ResNet. """


class PreActBlock(nn.Module):
    r""""Pre-activation Resnet Block used in ResNet-18"""

    def __init__(self, in_channels, out_channels, stride=1):
        super(PreActBlock, self).__init__()

        self.bn1 = nn.BatchNorm2d(in_channels)
        self.conv1 = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=3,
            stride=stride,
            padding=1,
            bias=False,
        )
        self.bn2 = nn.BatchNorm2d(out_channels)
        self.conv2 = nn.Conv2d(
            out_channels, out_channels, kernel_size=3, stride=1, padding=1, bias=False
        )

        if stride != 1 or in_channels != out_channels:
            self.shortcut = nn.Sequential(
                nn.Conv2d(
                    in_channels, out_channels, kernel_size=1, stride=stride, bias=False
                )
            )

    def forward(self, x):
        out = F.relu(self.bn1(x))
        shortcut = self.shortcut(out) if hasattr(self, "shortcut") else x
        out = self.conv1(out)
        out = self.conv2(F.relu(self.bn2(out)))
        return out + shortcut


class ResNet18CIFAR10(nn.Module):
    """ResNet implementation for CIFAR-10

    The ResNet structure defined in :cite:`he2016deep` and :cite:`he2016identity`.

    For CIFAR-10 dataset, the ResNet are configured to have 6n+2 layers where fixing n={3,5,7,9}
    gives ResNet-20,32,44,56 seperately. The input image is assumed to have a shape of 32*32 pixels.

    Args:
        layers (:obj:`list` of :obj:`int`): List of resnet blocks per layer. Must contain 4 elements.
        num_classes (int): Number of output classes. Default: ``1000``
    """

    def __init__(self, layers, num_classes=1000):
        super(ResNet18CIFAR10, self).__init__()
        self.prep = nn.Sequential(
            nn.Conv2d(3, 64, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(64),
            nn.ReLU(),
        )

        self.layers = nn.Sequential(
            self._make_layer(64, 64, layers[0], stride=1),
            self._make_layer(64, 128, layers[1], stride=2),
            self._make_layer(128, 256, layers[2], stride=2),
            self._make_layer(256, 256, layers[3], stride=2),
        )

        self.classifier = nn.Linear(512, num_classes)

    def _make_layer(self, in_channels, out_channels, num_blocks, stride):

        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for stride in strides:
            layers.append(
                PreActBlock(
                    in_channels=in_channels, out_channels=out_channels, stride=stride
                )
            )
            in_channels = out_channels

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.prep(x)

        x = self.layers(x)

        x_avg = F.adaptive_avg_pool2d(x, (1, 1))
        x_avg = x_avg.view(x_avg.size(0), -1)

        x_max = F.adaptive_max_pool2d(x, (1, 1))
        x_max = x_max.view(x_max.size(0), -1)

        x = torch.cat([x_avg, x_max], dim=-1)

        x = self.classifier(x)

        return x


def resnet18_bkj(num_classes):
    """Constructs a ResNet-18 model from DAWN.

    This
    `implementation <https://github.com/bkj/basenet/blob/49b2b61e5b9420815c64227c5a10233267c1fb14/examples/cifar10.py>`_
    comes from which gives results in
    `DAWNBench <https://github.com/stanford-futuredata/dawn-bench-entries/blob/master/CIFAR10/train/basenet.json>`_.

    Args:
        num_classes (int): The number of output classes.
    """
    model = ResNet18CIFAR10([2, 2, 2, 2], num_classes=num_classes)
    return model


def get_resnet_model(model, version, dtype, num_classes=1000, use_cuda=False):
    """ Create a resnet model

    Args:
        model (str): The name of the model, e.g. `resnet18`
        version (int): The resnet version to use, `1`or `2`
        num_classes (int): The number of output classes. Default: `1000`
        use_cuda (bool): Whether to train on the GPU or not. Default: `False`

    Returns:
        A `torch.nn.Module` Resnet Model
    """
    if model == "resnet18":
        model = resnet18_bkj(num_classes)
    elif model in ["resnet20", "resnet32", "resnet44", "resnet56", "resnet110"]:
        resnet_size = int(model[len("resnet") :])
        model = ResNetCIFAR(resnet_size, False, 10, version=version)
    else:
        raise NotImplementedError("{}_{} is not implemented.".format(model, version))

    model = convert_dtype(dtype, model)
    if use_cuda:
        model = model.cuda()
    return model



# Taken from https://github.com/NVIDIA/DeepLearningExamples/tree/master/PyTorch/Translation/GNMT
import torch

from mlbench_core.dataset.translation.pytorch.config import BOS, EOS


class SequenceGenerator:
    """
    Generator for the autoregressive inference with beam search decoding.
    """

    def __init__(
        self,
        model,
        beam_size=5,
        max_seq_len=100,
        len_norm_factor=0.6,
        len_norm_const=5,
        cov_penalty_factor=0.1,
    ):
        """
        Constructor for the SequenceGenerator.

        Beam search decoding supports coverage penalty and length
        normalization. For details, refer to Section 7 of the GNMT paper
        (https://arxiv.org/pdf/1609.08144.pdf).

        Args:
            model: model which implements generate method
            beam_size (int): decoder beam size
            max_seq_len (int):  maximum decoder sequence length
            len_norm_factor (float): length normalization factor
            len_norm_const (int): length normalization constant
            cov_penalty_factor (float): coverage penalty factor
        """
        self.model = model
        self.beam_size = beam_size
        self.max_seq_len = max_seq_len
        self.len_norm_factor = len_norm_factor
        self.len_norm_const = len_norm_const
        self.cov_penalty_factor = cov_penalty_factor

        self.batch_first = self.model.batch_first

    def greedy_search(self, batch_size, initial_input, initial_context=None):
        """
        Greedy decoder.

        Args:
            batch_size (int): decoder batch size
            initial_input (torch.tensor): initial input, usually tensor of BOS tokens
            initial_context (torch.tensor):  initial context, usually [encoder_context,
                src_seq_lengths, None]

        Returns: (translation, lengths, counter)
            translation: (batch_size, max_seq_len) - indices of target tokens
            lengths: (batch_size) - lengths of generated translations
            counter: number of iterations of the decoding loop
        """
        device = initial_input.device
        max_seq_len = self.max_seq_len

        translation = torch.zeros(
            batch_size, max_seq_len, dtype=torch.int64, device=device
        )
        lengths = torch.ones(batch_size, dtype=torch.int64, device=device)
        active = torch.arange(0, batch_size, dtype=torch.int64, device=device)
        base_mask = torch.arange(0, batch_size, dtype=torch.int64, device=device)

        translation[:, 0] = BOS
        words, context = initial_input, initial_context

        if self.batch_first:
            word_view = (-1, 1)
            ctx_batch_dim = 0
        else:
            word_view = (1, -1)
            ctx_batch_dim = 1

        counter = 0
        for idx in range(1, max_seq_len):
            if not len(active):
                break
            counter += 1

            words = words.view(word_view)
            output = self.model.generate(words, context, 1)
            words, logprobs, attn, context = output
            words = words.view(-1)

            translation[active, idx] = words
            lengths[active] += 1

            terminating = words == EOS

            if terminating.any():
                not_terminating = ~terminating

                mask = base_mask[: len(active)]
                mask = mask.masked_select(not_terminating)
                active = active.masked_select(not_terminating)

                words = words[mask]
                context[0] = context[0].index_select(ctx_batch_dim, mask)
                context[1] = context[1].index_select(0, mask)
                context[2] = context[2].index_select(1, mask)

        return translation, lengths, counter

    def beam_search(self, batch_size, initial_input, initial_context=None):
        """
        Beam Search decoder.

        Args:
            batch_size (int): decoder batch size
            initial_input (torch.tensor): initial input, usually tensor of BOS tokens
            initial_context (torch.tensor):  initial context, usually [encoder_context,
                src_seq_lengths, None]

        Returns: (translation, lengths, counter)
            translation: (batch_size, max_seq_len) - indices of target tokens
            lengths: (batch_size) - lengths of generated translations
            counter: number of iterations of the decoding loop
        """
        device = initial_input.device
        beam_size = self.beam_size
        norm_const = self.len_norm_const
        norm_factor = self.len_norm_factor
        max_seq_len = self.max_seq_len
        cov_penalty_factor = self.cov_penalty_factor

        translation = torch.zeros(
            batch_size * beam_size, max_seq_len, dtype=torch.int64, device=device
        )
        lengths = torch.ones(batch_size * beam_size, dtype=torch.int64, device=device)
        scores = torch.zeros(batch_size * beam_size, dtype=torch.float32, device=device)
        active = torch.arange(
            0, batch_size * beam_size, dtype=torch.int64, device=device
        )
        base_mask = torch.arange(
            0, batch_size * beam_size, dtype=torch.int64, device=device
        )
        global_offset = torch.arange(
            0, batch_size * beam_size, beam_size, device=device, dtype=torch.int64
        )
        eos_beam_fill = torch.tensor(
            [0] + (beam_size - 1) * [float("-inf")], dtype=torch.float32, device=device
        )

        translation[:, 0] = BOS

        words, context = initial_input, initial_context

        if self.batch_first:
            word_view = (-1, 1)
            ctx_batch_dim = 0
            attn_query_dim = 1
        else:
            word_view = (1, -1)
            ctx_batch_dim = 1
            attn_query_dim = 0

        # replicate context
        if self.batch_first:
            # context[0] (encoder state): (batch, seq, feature)
            _, seq, feature = context[0].shape
            context[0].unsqueeze_(1)
            context[0] = context[0].expand(-1, beam_size, -1, -1)
            context[0] = (
                context[0].contiguous().view(batch_size * beam_size, seq, feature)
            )
            # context[0]: (batch * beam, seq, feature)
        else:
            # context[0] (encoder state): (seq, batch, feature)
            seq, _, feature = context[0].shape
            context[0].unsqueeze_(2)
            context[0] = context[0].expand(-1, -1, beam_size, -1)
            context[0] = (
                context[0].contiguous().view(seq, batch_size * beam_size, feature)
            )
            # context[0]: (seq, batch * beam,  feature)

        # context[1] (encoder seq length): (batch)
        context[1].unsqueeze_(1)
        context[1] = context[1].expand(-1, beam_size)
        context[1] = context[1].contiguous().view(batch_size * beam_size)
        # context[1]: (batch * beam)

        accu_attn_scores = torch.zeros(
            batch_size * beam_size, seq, dtype=torch.float32, device=device
        )

        counter = 0
        for idx in range(1, self.max_seq_len):
            if not len(active):
                break
            counter += 1

            eos_mask = words == EOS
            eos_mask = eos_mask.view(-1, beam_size)

            terminating, _ = eos_mask.min(dim=1)

            lengths[active[~eos_mask.view(-1)]] += 1

            output = self.model.generate(words, context, beam_size)
            words, logprobs, attn, context = output

            attn = attn.float().squeeze(attn_query_dim)
            attn = attn.masked_fill(eos_mask.view(-1).unsqueeze(1), 0)
            accu_attn_scores[active] += attn

            # words: (batch, beam, k)
            words = words.view(-1, beam_size, beam_size)
            words = words.masked_fill(eos_mask.unsqueeze(2), EOS)

            # logprobs: (batch, beam, k)
            logprobs = logprobs.float().view(-1, beam_size, beam_size)

            if eos_mask.any():
                logprobs[eos_mask] = eos_beam_fill

            active_scores = scores[active].view(-1, beam_size)
            # new_scores: (batch, beam, k)
            new_scores = active_scores.unsqueeze(2) + logprobs

            if idx == 1:
                new_scores[:, 1:, :].fill_(float("-inf"))

            new_scores = new_scores.view(-1, beam_size * beam_size)
            # index: (batch, beam)
            _, index = new_scores.topk(beam_size, dim=1)
            source_beam = index / beam_size

            new_scores = new_scores.view(-1, beam_size * beam_size)
            best_scores = torch.gather(new_scores, 1, index)
            scores[active] = best_scores.view(-1)

            words = words.view(-1, beam_size * beam_size)
            words = torch.gather(words, 1, index)

            # words: (1, batch * beam)
            words = words.view(word_view)

            offset = global_offset[: source_beam.shape[0]]
            source_beam += offset.unsqueeze(1)

            translation[active, :] = translation[active[source_beam.view(-1)], :]
            translation[active, idx] = words.view(-1)

            lengths[active] = lengths[active[source_beam.view(-1)]]

            context[2] = context[2].index_select(1, source_beam.view(-1))

            if terminating.any():
                not_terminating = ~terminating
                not_terminating = not_terminating.unsqueeze(1)
                not_terminating = not_terminating.expand(-1, beam_size).contiguous()

                normalization_mask = active.view(-1, beam_size)[terminating]

                # length normalization
                norm = lengths[normalization_mask].float()
                norm = (norm_const + norm) / (norm_const + 1.0)
                norm = norm ** norm_factor

                scores[normalization_mask] /= norm

                # coverage penalty
                penalty = accu_attn_scores[normalization_mask]
                penalty = penalty.clamp(0, 1)
                penalty = penalty.log()
                penalty[penalty == float("-inf")] = 0
                penalty = penalty.sum(dim=-1)

                scores[normalization_mask] += cov_penalty_factor * penalty

                mask = base_mask[: len(active)]
                mask = mask.masked_select(not_terminating.view(-1))

                words = words.index_select(ctx_batch_dim, mask)
                context[0] = context[0].index_select(ctx_batch_dim, mask)
                context[1] = context[1].index_select(0, mask)
                context[2] = context[2].index_select(1, mask)

                active = active.masked_select(not_terminating.view(-1))

        scores = scores.view(batch_size, beam_size)
        _, idx = scores.max(dim=1)

        translation = translation[idx + global_offset, :]
        lengths = lengths[idx + global_offset]

        return translation, lengths, counter



"""Tests for `mlbench_core.utils.pytorch.distributed`"""
import torch
from mlbench_core.utils.pytorch.distributed import pack_tensors, unpack_tensors


def test_pack_tensors():
    tensors = [torch.rand(2, 2), torch.rand(2, 2)]

    flattened = [y for x in tensors for y in x.view(-1)]

    vec, indices, sizes = pack_tensors(tensors)

    assert vec.tolist() == flattened
    assert indices == [0, 4, 8]
    assert sizes == [(2, 2), (2, 2)]


def test_unpack_tensors():
    tensors = [torch.rand(2, 2), torch.rand(2, 2)]
    vec, indices, sizes = pack_tensors(tensors)

    unpacked = unpack_tensors(vec, indices, sizes)

    assert all((x == y).all() for x, y in zip(tensors, unpacked))



r"""Control flow for pytorch applications."""
import logging

from mlbench_core.utils import AverageMeter, Tracker
from mlbench_core.utils.pytorch.distributed import global_average
from mlbench_core.utils.pytorch.helpers import iterate_dataloader

import torch
import torch.distributed as dist

logger = logging.getLogger("mlbench")

LOG_EVERY_N_BATCHES = 25


def _record_train_batch_stats(
    batch_idx, loss, output, target, metrics, tracker, num_batches_per_device_train
):
    r"""Record the stats in a training batch.

    Args:
        batch_idx (int): The id of the current batch
        loss (float): The loss of the batch
        output (:obj:`torch.Tensor`): The model output
        target (:obj:`torch.Tensor`): The labels for the current batch
        metrics (list): List of metrics to track
        tracker (`obj`:mlbench_core.utils.Tracker): Tracker object to use.
        num_batches_per_device_train (int): Number of batches per train epoch
    """
    progress = batch_idx / num_batches_per_device_train
    progress += tracker.current_epoch

    log_to_api = (
        batch_idx % LOG_EVERY_N_BATCHES == 0
        or batch_idx == num_batches_per_device_train
    )

    if tracker:
        tracker.record_loss(loss, output.size()[0], log_to_api=log_to_api)

    # Compute metrics for one batch
    for metric in metrics:
        metric_value = metric(loss, output, target).item()

        if tracker:
            tracker.record_metric(
                metric, metric_value, output.size()[0], log_to_api=log_to_api
            )

    status = "Epoch {:5.2f} Batch {:4}: ".format(progress, batch_idx)

    logger.info(status + str(tracker))


def train_round(
    dataloader,
    model,
    optimizer,
    loss_function,
    metrics,
    scheduler,
    dtype,
    schedule_per="epoch",
    transform_target_type=None,
    use_cuda=False,
    max_batch_per_epoch=None,
    tracker=None,
):
    """ Performs max_batch_per_epoch batches of training (or full trainset if
    not specified)

    Args:
        dataloader (:obj:`torch.utils.data.DataLoader`): The train set
        model (`obj`:torch.nn.Module): The model to train
        optimizer (`obj`:torch.optim): The optimizer
        loss_function (`obj`:torch.nn.Module): The loss function
        metrics (list): List of metrics to track
        scheduler (`obj`:torch.optim.lr_scheduler): Learning Rate scheduler
        dtype (str): The datatype to use, one of `fp32`or `fp64`
        scheduler_per (str): Learning Rate scheduler mode, one of `batch` or `epoch`
        transform_target_type (str): Datatype to convert data to, default: `None`
        use_cuda (bool): Whether to use GPU for training, default: `False`
        max_batch_per_epoch (int): Maximum number of batches tot rain for per epoch,
                                   default: `None` (all batches)
        tracker (`obj`:mlbench_core.utils.Tracker): Tracker object to use.
    """
    model.train()

    if tracker:
        tracker.train()

    data_iter = iterate_dataloader(
        dataloader, dtype, max_batch_per_epoch, use_cuda, transform_target_type
    )

    num_batches_per_device_train = len(dataloader)

    for batch_idx, (data, target) in enumerate(data_iter):
        if tracker:
            tracker.batch_start()

        # Clear gradients in the optimizer.
        optimizer.zero_grad()
        if tracker:
            tracker.record_batch_step("init")

        # Compute the output
        output = model(data)
        if tracker:
            tracker.record_batch_step("fwd_pass")

        # Compute the loss
        loss = loss_function(output, target)
        if tracker:
            tracker.record_batch_step("comp_loss")

        # Backprop
        loss.backward()
        if tracker:
            tracker.record_batch_step("backprop")

        # Aggregate gradients/parameters from all workers and apply updates to model
        optimizer.step()
        if tracker:
            tracker.record_batch_step("opt_step")

        if schedule_per == "batch":
            scheduler.step()

        if tracker:
            tracker.batch_end()

        _record_train_batch_stats(
            batch_idx,
            loss.item(),
            output,
            target,
            metrics,
            tracker,
            num_batches_per_device_train,
        )
    if schedule_per == "epoch":
        scheduler.step()


def _validate(
    dataloader,
    model,
    loss_function,
    metrics,
    dtype,
    transform_target_type=None,
    use_cuda=False,
    max_batch_per_epoch=None,
):
    """Evaluate the model on the test dataset.

    Args:
        dataloader (:obj:`torch.utils.data.DataLoader`): The validation set
        model (`obj`:torch.nn.Module): The model to train
        loss_function (`obj`:torch.nn.Module): The loss function
        metrics (list): List of metrics to track
        dtype (str): The datatype to use, one of `fp32`or `fp64`
        transform_target_type (str): Datatype to convert data to, default: `None`
        use_cuda (bool): Whether to use GPU for training, default: `False`
        max_batch_per_epoch (int): Maximum number of batches tot rain for per epoch,
                                   default: `None` (all batches)
        """
    # Initialize the accumulators for loss and metrics
    losses = AverageMeter()
    for metric in metrics:
        metric.reset()

    # Each worker computer their own losses and metrics
    with torch.no_grad():
        data_iter = iterate_dataloader(
            dataloader, dtype, max_batch_per_epoch, use_cuda, transform_target_type
        )

        for data, target in data_iter:
            # Inference
            output = model(data)

            # Compute loss
            loss = loss_function(output, target)

            # Update loss
            losses.update(loss.item(), data.size(0))

            # Update metrics
            for metric in metrics:
                metric_value = metric(loss, output, target)
                metric.update(metric_value, data.size(0))

    # Aggregate metrics and loss for all workers
    metrics_averages = {metric: metric.average().item() for metric in metrics}
    loss_average = global_average(losses.sum, losses.count).item()
    return metrics_averages, loss_average


def validation_round(
    dataloader,
    model,
    loss_function,
    metrics,
    run_id,
    rank,
    dtype,
    transform_target_type=None,
    use_cuda=False,
    max_batch_per_epoch=None,
    tracker=None,
):
    """ Handles one full iteration of validation on the whole validation set.

    Args:
        dataloader (:obj:`torch.utils.data.DataLoader`): The validation set
        model (`obj`:torch.nn.Module): The model to train
        loss_function (`obj`:torch.nn.Module): The loss function
        metrics (list): List of metrics to track
        run_id (int): The id of the current run
        rank (int): The rank of the current worker node
        dtype (str): The datatype to use, one of `fp32`or `fp64`
        transform_target_type (str): Datatype to convert data to, default: `None`
        use_cuda (bool): Whether to use GPU for training, default: `False`
        max_batch_per_epoch (int): Maximum number of batches tot rain for per epoch,
                                   default: `None` (all batches)
        tracker (`obj`:mlbench_core.utils.Tracker): Tracker object to use. Will be
                                                    created if not supplied
    """
    model.eval()

    if tracker:
        tracker.validation()

        tracker.validation_start()

    metrics_values, loss = _validate(
        dataloader,
        model,
        loss_function,
        metrics,
        dtype,
        transform_target_type,
        use_cuda,
        max_batch_per_epoch,
    )
    if tracker:
        tracker.validation_end()

    if len(metrics_values) > 0:
        # Save
        if tracker:
            for metric, value in metrics_values.items():
                tracker.record_metric(metric, value, log_to_api=True)

                global_metric_value = global_average(value, 1).item()

                if rank == 0:
                    tracker.record_stat(
                        "global_{}".format(metric.name),
                        global_metric_value,
                        log_to_api=True,
                    )

        if rank == 0 and tracker:
            logger.info(
                "{} for rank {}:(best epoch {}, current epoch {}): {:.3f}".format(
                    tracker.primary_metric.name,
                    tracker.rank,
                    tracker.best_epoch,
                    tracker.current_epoch,
                    tracker.best_metric_value,
                )
            )
    else:
        if rank == 0:
            logger.info("Validation loss={:.3f}".format(loss))

    if tracker:
        tracker.record_loss(loss, log_to_api=True)

        global_loss = global_average(loss, 1).item()

        if rank == 0:
            tracker.record_stat("global_loss", global_loss, log_to_api=True)

    return tracker.is_best() if tracker else False


class TrainValidation(object):
    r"""Train and validate a model.

    Args:
        model (:obj:`torch.nn.Module`): a pytorch model to be trained and validated.
        optimizer (:obj:`torch.optim.Optimizer`): an optimizer for the given model.
        loss_function (:obj:`torch.nn.modules.loss._Loss`): loss function.
        metrics (:obj:`list` of :obj:`mlbench_core.evaluation.pytorch.*`): metrics like TopKAccuracy.
        scheduler (:obj:`mlbench_core.lr_scheduler.pytorch.lr.*`): a scheduler for hyperparameters.
        batch_size (int): The size of batches provided by the dataloader
        train_epochs (int): The number of epochs to train for
        rank (int): The rank of the current workers
        world_size (int): The total number of workers
        run_id (str): The id of the current run
        dtype (str): The datatype to use for the dataloader data
        validate (bool): Whether to run validation on the val dataset. Default: `True`
        schedule_per (str): When to perform a step for the lr scheduler, one of
            `epoch` or `batch`. Default: `epoch`
        checkpoint (:obj:`Checkpointer`): Class that handles checkpointing. Default: `None`
        transform_target_type (str): dtype to transform the target to. Not used. Default: `None`
        average_models (bool): Whether to average models together. Default: `False`
        use_cuda (bool): Whether to train on GPU or not. Default: `False`
        max_batch_per_epoch (int): Maximum number of batches per epoch. Whole dataset
            is used if not specified. Default: `None`
        tracker (:obj:`mlbench_core.utils.Tracker`): Tracker for the controlflow. Default: `None`
    """

    def __init__(
        self,
        model,
        optimizer,
        loss_function,
        metrics,
        scheduler,
        batch_size,
        train_epochs,
        rank,
        world_size,
        run_id,
        dtype,
        validate=True,
        schedule_per="epoch",
        checkpoint=None,
        transform_target_type=None,
        average_models=False,
        use_cuda=False,
        max_batch_per_epoch=None,
        tracker=None,
    ):
        self.batch_size = batch_size
        self.train_epochs = train_epochs
        self.model = model
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.schedule_per = schedule_per
        self.perform_validation = validate
        self.checkpoint = checkpoint
        self.model = model
        self.optimizer = optimizer
        self.loss_function = loss_function
        self.metrics = metrics
        self.scheduler = scheduler
        self.batch_size = batch_size
        self.rank = rank
        self.run_id = run_id
        self.dtype = dtype
        self.schedule_per = schedule_per
        self.transform_target_type = transform_target_type
        self.use_cuda = use_cuda
        self.max_batch_per_epoch = max_batch_per_epoch
        if tracker:
            self.tracker = tracker
        else:
            self.tracker = Tracker(metrics, run_id, rank)

    def _get_dataloader_stats(self, dataloader_train, dataloader_val):
        """ Sets the stats for the supplied dataloaders

        Args:
            dataloader_train (:obj:`torch.utils.data.DataLoader`): The train set
            dataloader_val (:obj:`torch.utils.data.DataLoader`): The validation set
        """
        self.num_batches_per_device_train = len(dataloader_train)
        self.num_batches_per_device_val = len(dataloader_val)

    def run(
        self,
        dataloader_train=None,
        dataloader_val=None,
        dataloader_train_fn=None,
        dataloader_val_fn=None,
        resume=False,
        repartition_per_epoch=False,
    ):
        """Execute training and (possibly) validation

        `dataloader_train` and `dataloader_train_fn` are mutually exclusive.
        `dataloader_val` and `dataloader_val_fn` are mutually exclusive.

        Args:
            dataloader_train (:obj:`torch.utils.data.DataLoader`): A dataloader for the train set.
                Default: `None`
            dataloader_val (:obj:`torch.utils.data.DataLoader`): A dataloader for the val set.
                Default: `None`
            dataloader_train_fn (:func:`Function`): A function returning a :obj:`torch.utils.data.DataLoader`
                for the train set. Default: `None`
            dataloader_val_fn (:func:`Function`): A function returning a :obj:`torch.utils.data.DataLoader`
                for the val set. Default: `None`
            resume (bool): Whether this is a resume of a previous run or not. Default: `False`
            repartition_per_epoch (bool): Whether to repartition the dataset again every epoch.
                Requires dataloader_train_fn and/or dataloader_val_fn to be set. Default: `False`
        """

        if not dataloader_train_fn and not dataloader_train:
            raise ValueError(
                "One of dataloader_train_fn or dataloader_train must be set"
            )

        if not dataloader_val_fn and not dataloader_val:
            raise ValueError("One of dataloader_val_fn or dataloader_val must be set")

        if dataloader_train_fn:
            dataloader_train = dataloader_train_fn()

        if dataloader_val_fn:
            dataloader_val = dataloader_val_fn()

        self._get_dataloader_stats(dataloader_train, dataloader_val)

        # define some parameters for training.
        logger.info(
            "There are {train_epochs} epochs, {num_batches} "
            "mini-batches per epoch (batch size: {batch_size}).".format(
                train_epochs=self.train_epochs,
                num_batches=self.num_batches_per_device_train,
                batch_size=self.batch_size,
            )
        )

        # Initialize Tracker or resume from checkpoint
        if resume:
            start_epoch = self.tracker.current_epoch + 1
        else:
            start_epoch = 0

        dist.barrier()
        for epoch in range(start_epoch, self.train_epochs):
            # Per epoch information.
            logger.info(
                "Current epoch : {} : lr={}".format(epoch, self.scheduler.get_lr())
            )

            train_round(
                dataloader_train,
                self.model,
                self.optimizer,
                self.loss_function,
                self.metrics,
                self.scheduler,
                self.dtype,
                self.schedule_per,
                self.transform_target_type,
                self.use_cuda,
                self.max_batch_per_epoch,
                self.tracker,
            )

            is_best = False
            if self.perform_validation:
                is_best = validation_round(
                    dataloader_val,
                    self.model,
                    self.loss_function,
                    self.metrics,
                    self.run_id,
                    self.rank,
                    self.dtype,
                    self.transform_target_type,
                    self.use_cuda,
                    self.max_batch_per_epoch,
                    self.tracker,
                )

            if self.checkpoint:
                self.checkpoint.save(
                    self.tracker,
                    self.model,
                    self.optimizer,
                    self.scheduler,
                    self.tracker.current_epoch,
                    is_best,
                )

            # Shuffle the dataset across nodes
            if repartition_per_epoch:
                if dataloader_train_fn:
                    dataloader_train = dataloader_train_fn()

                if dataloader_val_fn:
                    dataloader_val = dataloader_val_fn()

                self._get_dataloader_stats(dataloader_train, dataloader_val)

            self.tracker.epoch_end()



import bz2
import os
import sys
import tarfile
from urllib.request import urlretrieve


def progress_download(url, dest):
    """ Downloads a file from `url` to `dest` and shows progress

    Args:
        url (src): Url to retrieve file from
        dest (src): Destination file
    """

    def _progress(count, block_size, total_size):
        percentage = float(count * block_size) / float(total_size) * 100.0
        if percentage % 25 == 0:
            sys.stdout.write(
                "\r>> Downloading %s %.1f%%" % (os.path.basename(dest), percentage)
            )
            sys.stdout.flush()

    urlretrieve(url, dest, _progress)
    print("\nDownloaded {} to {}\n".format(url, dest))


def extract_bz2_file(source, dest, delete=True):
    """ Extracts a bz2 archive

    Args:
        source (str): Source file (must have .bz2 extension)
        dest (str): Destination file
        delete (bool): Delete compressed file after decompression

    """
    assert source.endswith(".bz2"), "Extracting non bz2 archive"

    with open(dest, "wb") as d, open(source, "rb") as s:
        decompressor = bz2.BZ2Decompressor()
        for data in iter(lambda: s.read(1000 * 1024), b""):
            d.write(decompressor.decompress(data))

    if delete:
        os.remove(source)


def compress_to_bz2_file(source, delete=True):
    """ Extracts a bz2 archive

    Args:
        source (str): Source file to compress
        delete (bool): Delete un-compressed file
    """

    dest = source + ".bz2"
    with open(source, "rb") as s, open(dest, "wb") as d:
        compressor = bz2.BZ2Compressor()
        for data in iter(lambda: s.read(1000 * 1024), b""):
            d.write(compressor.compress(data))

    if delete:
        os.remove(source)


def maybe_download_and_extract_bz2(root, file_name, data_url):
    """ Downloads file from given URL and extracts if bz2

    Args:
        root (str): The root directory
        file_name (str): File name to download to
        data_url (str): Url of data
    """
    if not os.path.exists(root):
        os.makedirs(root)

    file_path = os.path.join(root, file_name)

    # Download file if not present
    if len([x for x in os.listdir(root) if x == file_name]) == 0:
        progress_download(data_url, file_path)

    # Extract downloaded file if compressed
    if file_name.endswith(".bz2"):
        file_basename = os.path.splitext(file_name)[0]
        extracted_fpath = os.path.join(root, file_basename)

        # Extract file
        extract_bz2_file(file_path, extracted_fpath, delete=True)
        file_path = extracted_fpath
    return file_path


def maybe_download_and_extract_tar_gz(root, file_name, data_url):
    if not os.path.exists(root):
        os.makedirs(root)

    file_path = os.path.join(root, file_name)

    # Download file if not present
    if len([x for x in os.listdir(root) if x == file_name]) == 0:
        progress_download(data_url, file_path)

    if file_name.endswith(".tar.gz"):
        with tarfile.open(file_path, "r:gz") as tar:
            dirs = [member for member in tar.getmembers()]
            tar.extractall(path=root, members=dirs)
